@@ -79,7 +79,7 @@ long long read_binary(char *src, int size)
 
 long long read_number(char *s, int size)
 {
-	if(*s == -128)
+	if(*((signed char *) s) == -128)
 		return read_binary(s + 1, size - 1);
 	else
 		return read_octal(s, size);
@@ -121,8 +121,9 @@ char *read_long_string(int size, int skip)
 		int expected = size > 512 ? 512 : size;
 
 		res = read_bytes(STDIN_FILENO, buffer, 512);
-		if(res == FALSE) {
-			ERROR("Unexpected EOF (end of file), the tarfile appears to be truncated or corrupted\n");
+		if(res < 512) {
+			if(res != -1)
+				ERROR("Unexpected EOF (end of file), the tarfile appears to be truncated or corrupted\n");
 			free(name);
 			return NULL;
 		}
@@ -244,6 +245,7 @@ static struct inode_info *new_inode(struct tar_file *tar_file)
 	inode->inode = SQUASHFS_INVALID_BLK;
 	inode->nlink = 1;
 	inode->inode_number = 0;
+	inode->pseudo = NULL;
 	inode->dummy_root_dir = FALSE;
 	inode->tarfile = TRUE;
 
@@ -265,6 +267,21 @@ static struct inode_info *new_inode(struct tar_file *tar_file)
 }
 
 
+static struct inode_info *copy_inode(struct inode_info *source)
+{
+	struct inode_info *inode;
+	int bytes = S_ISLNK(source->buf.st_mode) ? strlen(source->symlink) + 1 : 0;
+
+	inode = malloc(sizeof(struct inode_info) + bytes);
+	if(inode == NULL)
+		MEM_ERROR();
+
+	memcpy(inode, source, sizeof(struct inode_info) + bytes);
+
+	return inode;
+}
+
+
 static void fixup_tree(struct dir_info *dir)
 {
 	struct dir_ent *entry;
@@ -283,6 +300,7 @@ static void fixup_tree(struct dir_info *dir)
 			buf.st_dev = 0;
 			buf.st_ino = 0;
 			entry->inode = lookup_inode(&buf);
+			entry->inode->tar_file = NULL;
 			entry->inode->tarfile = TRUE;
 		}
 		
@@ -377,6 +395,8 @@ static struct dir_info *add_tarfile(struct dir_info *sdir, char *source,
 					goto failed_early;
 			}
 		}
+
+		free(name);
 	} else {
 		/*
 		 * No matching name found.
@@ -395,12 +415,14 @@ static struct dir_info *add_tarfile(struct dir_info *sdir, char *source,
 			if(S_ISDIR(tarfile->buf.st_mode)) {
 				add_dir_entry(entry, NULL, new_inode(tarfile));
 				dir->directory_count ++;
-			} else {
-				struct inode_info *new = link ? link : new_inode(tarfile);
-				add_dir_entry(entry, NULL, new);
+			} else if (link == FALSE) {
+				add_dir_entry(entry, NULL, new_inode(tarfile));
 				if(S_ISREG(tarfile->buf.st_mode))
 					*dir_ent = entry;
-			}
+			} else if(no_hardlinks)
+				add_dir_entry(entry, NULL, copy_inode(link));
+			else
+				add_dir_entry(entry, NULL, link);
 		} else {
 			subpath = subpathname(entry);
 			sub = add_tarfile(NULL, source, subpath, tarfile, new, depth + 1, dir_ent, link);
@@ -692,8 +714,9 @@ int read_pax_header(struct tar_file *file, long long st_size)
 		MEM_ERROR();
 
 	res = read_bytes(STDIN_FILENO, data, size);
-	if(res == FALSE) {
-		ERROR("Unexpected EOF (end of file), the tarfile appears to be truncated or corrupted\n");
+	if(res < size) {
+		if(res != -1)
+			ERROR("Unexpected EOF (end of file), the tarfile appears to be truncated or corrupted\n");
 		free(data);
 		return FALSE;
 	}
@@ -946,8 +969,9 @@ struct file_map *read_sparse_headers(struct tar_file *file, struct short_sparse_
 
 	while(isextended) {
 		res = read_bytes(STDIN_FILENO, &long_header, 512);
-		if(res == FALSE) {
-			ERROR("Unexpected EOF (end of file), the tarfile appears to be truncated or corrupted\n");
+		if(res < 512) {
+			if(res != -1)
+				ERROR("Unexpected EOF (end of file), the tarfile appears to be truncated or corrupted\n");
 			goto failed;
 		}
 
@@ -1011,8 +1035,9 @@ struct file_map *read_sparse_map(struct tar_file *file, int *entries)
 	int atoffset = TRUE, i = 0;
 
 	res = read_bytes(STDIN_FILENO, buffer, 512);
-	if(res == FALSE) {
-		ERROR("Unexpected EOF (end of file), the tarfile appears to be truncated or corrupted\n");
+	if(res < 512) {
+		if(res != -1)
+			ERROR("Unexpected EOF (end of file), the tarfile appears to be truncated or corrupted\n");
 		goto failed;
 	}
 
@@ -1044,8 +1069,9 @@ struct file_map *read_sparse_map(struct tar_file *file, int *entries)
 
 			memmove(buffer, src, size);
 			res = read_bytes(STDIN_FILENO, buffer + size, 512);
-			if(res == FALSE) {
-				ERROR("Unexpected EOF (end of file), the tarfile appears to be truncated or corrupted\n");
+			if(res < 512) {
+				if(res != -1)
+					ERROR("Unexpected EOF (end of file), the tarfile appears to be truncated or corrupted\n");
 				goto failed;
 			}
 
@@ -1119,13 +1145,15 @@ static struct tar_file *read_tar_header(int *status)
 
 again:
 	res = read_bytes(STDIN_FILENO, &header, 512);
-	if(res == FALSE) {
-		ERROR("Unexpected EOF (end of file), the tarfile appears to be truncated or corrupted\n");
+	if(res < 512) {
+		if(res != -1)
+			ERROR("Unexpected EOF (end of file), the tarfile appears to be truncated or corrupted\n");
 		goto failed;
 	}
 
 	if(all_zero(&header)) {
 		*status = TAR_EOF;
+		free(file);
 		return NULL;
 	}
 
@@ -1386,8 +1414,9 @@ ignored:
 
 		while(size > 0) {
 			res = read_bytes(STDIN_FILENO, &header, 512);
-			if(res == FALSE) {
-				ERROR("Unexpected EOF (end of file), the tarfile appears to be truncated or corrupted\n");
+			if(res < 512) {
+				if(res != -1)
+					ERROR("Unexpected EOF (end of file), the tarfile appears to be truncated or corrupted\n");
 				goto failed;
 			}
 			size -= 512;
@@ -1443,6 +1472,7 @@ void read_tar_file()
 								 >> block_log);
 
 		file_buffer->cache = NULL;
+		file_buffer->fragment = FALSE;
 		file_buffer->tar_file = tar_file;
 		file_buffer->sequence = seq ++;
 		seq_queue_put(to_main, file_buffer);
@@ -1462,13 +1492,15 @@ squashfs_inode process_tar_file(int progress)
 	struct dir_info *new;
 	struct dir_ent *dir_ent;
 	struct tar_file *tar_file;
+	struct file_buffer *file_buffer;
 
 	queue_put(to_reader, NULL);
 	set_progressbar_state(progress);
 
 	while(1) {
 		struct inode_info *link = NULL;
-		struct file_buffer *file_buffer = seq_queue_get(to_main);
+
+		file_buffer = seq_queue_get(to_main);
 		if(file_buffer->tar_file == NULL)
 			break;
 
@@ -1512,18 +1544,27 @@ squashfs_inode process_tar_file(int progress)
 				update_info(dir_ent);
 				tar_file->file = write_file(dir_ent, &duplicate_file);
 				dir_ent->inode->read = TRUE;
-				INFO("file %s, uncompressed size %lld " "bytes %s\n", tar_file->pathname,
+				INFO("file %s, uncompressed size %lld bytes %s\n", tar_file->pathname,
 					(long long) tar_file->buf.st_size, duplicate_file ?  "DUPLICATE" : "");
 			}
 
-			if(link)
-				link->nlink ++;
+			if(link) {
+				if(no_hardlinks)
+					INFO("file %s, uncompressed size %lld bytes DUPLICATE\n", tar_file->pathname,
+						(long long) link->buf.st_size);
+				else
+					link->nlink ++;
+				free(tar_file->pathname);
+				free(tar_file);
+			}
 		} else if(S_ISREG(tar_file->buf.st_mode))
 			skip_file(file_buffer->tar_file);
 
 		free(file_buffer);
 	}
 
+	free(file_buffer);
+
 	if(root_dir)
 		fixup_tree(root_dir);
 	else
