@@ -2,7 +2,7 @@
  * Unsquash a squashfs filesystem.  This is a highly compressed read only
  * filesystem.
  *
- * Copyright (c) 2009, 2010, 2013, 2019
+ * Copyright (c) 2009, 2010, 2013, 2019, 2021, 2022
  * Phillip Lougher <phillip@squashfs.org.uk>
  *
  * This program is free software; you can redistribute it and/or
@@ -24,20 +24,37 @@
 
 #include "unsquashfs.h"
 #include "squashfs_compat.h"
+#include "compressor.h"
 
 static squashfs_fragment_entry_2 *fragment_table;
 static unsigned int *uid_table, *guid_table;
-static char *inode_table, *directory_table;
 static squashfs_operations ops;
+static int needs_sorting = FALSE;
 
-static void read_block_list(unsigned int *block_list, char *block_ptr, int blocks)
+
+static void read_block_list(unsigned int *block_list, long long start,
+					unsigned int offset, int blocks)
 {
+	int res;
+
 	TRACE("read_block_list: blocks %d\n", blocks);
 
 	if(swap) {
+		char *block_ptr = malloc(blocks * sizeof(unsigned int));
+		if(block_ptr == NULL)
+			MEM_ERROR();
+		res = read_inode_data(block_ptr, &start, &offset, blocks * sizeof(unsigned int));
+		if(res == FALSE)
+			EXIT_UNSQUASH("read_block_list: failed to read "
+				"inode index %lld:%d\n", start, offset);
 		SQUASHFS_SWAP_INTS_3(block_list, block_ptr, blocks);
-	} else
-		memcpy(block_list, block_ptr, blocks * sizeof(unsigned int));
+		free(block_ptr);
+	} else {
+		res = read_inode_data(block_list, &start, &offset, blocks * sizeof(unsigned int));
+		if(res == FALSE)
+			EXIT_UNSQUASH("read_block_list: failed to read "
+				"inode index %lld:%d\n", start, offset);
+	}
 }
 
 
@@ -71,20 +88,17 @@ static int read_fragment_table(long long *table_start)
 
 	fragment_table_index = malloc(length);
 	if(fragment_table_index  == NULL)
-		EXIT_UNSQUASH("read_fragment_table: failed to allocate "
-			"fragment table index\n");
+		MEM_ERROR();
 
 	fragment_table = malloc(bytes);
 	if(fragment_table == NULL)
-		EXIT_UNSQUASH("read_fragment_table: failed to allocate "
-			"fragment table\n");
+		MEM_ERROR();
 
 	if(swap) {
 		 unsigned int *sfragment_table_index = malloc(length);
 
 		if(sfragment_table_index == NULL)
-			EXIT_UNSQUASH("read_fragment_table: failed to allocate "
-				"fragment table index\n");
+			MEM_ERROR();
 
 		 res = read_fs_bytes(fd, sBlk.s.fragment_table_start,
 			length, sfragment_table_index);
@@ -157,28 +171,42 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 {
 	static union squashfs_inode_header_2 header;
 	long long start = sBlk.s.inode_table_start + start_block;
-	int bytes = lookup_entry(inode_table_hash, start);
-	char *block_ptr = inode_table + bytes + offset;
+	long long st = start;
+	unsigned int off = offset;
 	static struct inode i;
+	int res;
 
 	TRACE("read_inode: reading inode [%d:%d]\n", start_block,  offset);
 
-	if(bytes == -1)
-		EXIT_UNSQUASH("read_inode: inode table block %lld not found\n",
-			start); 
-
 	if(swap) {
 		squashfs_base_inode_header_2 sinode;
-		memcpy(&sinode, block_ptr, sizeof(header.base));
-		SQUASHFS_SWAP_BASE_INODE_HEADER_2(&header.base, &sinode,
-			sizeof(squashfs_base_inode_header_2));
+		res = read_inode_data(&sinode, &st, &off, sizeof(sinode));
+		if(res)
+			SQUASHFS_SWAP_BASE_INODE_HEADER_2(&header.base, &sinode,
+				sizeof(squashfs_base_inode_header_2));
 	} else
-		memcpy(&header.base, block_ptr, sizeof(header.base));
+		res = read_inode_data(&header.base, &st, &off, sizeof(header.base));
+
+	if(res == FALSE)
+		EXIT_UNSQUASH("read_inode: failed to read inode %lld:%d\n", st, off);
 
 	i.xattr = SQUASHFS_INVALID_XATTR;
+
+	if(header.base.uid >= sBlk.no_uids)
+		EXIT_UNSQUASH("File system corrupted - uid index in inode too large (uid: %u)\n", header.base.uid);
+
 	i.uid = (uid_t) uid_table[header.base.uid];
-	i.gid = header.base.guid == SQUASHFS_GUIDS ? i.uid :
-		(uid_t) guid_table[header.base.guid];
+
+	if(header.base.guid == SQUASHFS_GUIDS)
+		i.gid = i.uid;
+	else if(header.base.guid >= sBlk.no_guids)
+		EXIT_UNSQUASH("File system corrupted - gid index in inode too large (gid: %d)\n", header.base.guid);
+	else
+		i.gid = (uid_t) guid_table[header.base.guid];
+
+	if(header.base.inode_type < 1 || header.base.inode_type > 8)
+		EXIT_UNSQUASH("File system corrupted - invalid type in inode (type: %u)\n", header.base.inode_type);
+
 	i.mode = lookup_type[header.base.inode_type] | header.base.mode;
 	i.type = header.base.inode_type;
 	i.time = sBlk.s.mkfs_time;
@@ -190,12 +218,16 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 
 			if(swap) {
 				squashfs_dir_inode_header_2 sinode;
-				memcpy(&sinode, block_ptr, sizeof(header.dir));
-				SQUASHFS_SWAP_DIR_INODE_HEADER_2(&header.dir,
-					&sinode);
+				res = read_inode_data(&sinode, &start, &offset, sizeof(sinode));
+				if(res)
+					SQUASHFS_SWAP_DIR_INODE_HEADER_2(&header.dir,
+						&sinode);
 			} else
-				memcpy(&header.dir, block_ptr,
-					sizeof(header.dir));
+				res = read_inode_data(inode, &start, &offset, sizeof(*inode));
+
+			if(res == FALSE)
+				EXIT_UNSQUASH("read_inode: failed to read "
+					"inode %lld:%d\n", start, offset);
 
 			i.data = inode->file_size;
 			i.offset = inode->offset;
@@ -208,12 +240,16 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 
 			if(swap) {
 				squashfs_ldir_inode_header_2 sinode;
-				memcpy(&sinode, block_ptr, sizeof(header.ldir));
-				SQUASHFS_SWAP_LDIR_INODE_HEADER_2(&header.ldir,
-					&sinode);
+				res = read_inode_data(&sinode, &start, &offset, sizeof(sinode));
+				if(res)
+					SQUASHFS_SWAP_LDIR_INODE_HEADER_2(&header.ldir,
+						&sinode);
 			} else
-				memcpy(&header.ldir, block_ptr,
-					sizeof(header.ldir));
+				res = read_inode_data(inode, &start, &offset, sizeof(*inode));
+
+			if(res == FALSE)
+				EXIT_UNSQUASH("read_inode: failed to read "
+					"inode %lld:%d\n", start, offset);
 
 			i.data = inode->file_size;
 			i.offset = inode->offset;
@@ -226,11 +262,16 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 
 			if(swap) {
 				squashfs_reg_inode_header_2 sinode;
-				memcpy(&sinode, block_ptr, sizeof(sinode));
-				SQUASHFS_SWAP_REG_INODE_HEADER_2(inode,
-					&sinode);
+				res = read_inode_data(&sinode, &start, &offset, sizeof(sinode));
+				if(res)
+					SQUASHFS_SWAP_REG_INODE_HEADER_2(inode,
+						&sinode);
 			} else
-				memcpy(inode, block_ptr, sizeof(*inode));
+				res = read_inode_data(inode, &start, &offset, sizeof(*inode));
+
+			if(res == FALSE)
+				EXIT_UNSQUASH("read_inode: failed to read "
+					"inode %lld:%d\n", start, offset);
 
 			i.data = inode->file_size;
 			i.time = inode->mtime;
@@ -243,8 +284,9 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 				sBlk.s.block_log : i.data >>
 				sBlk.s.block_log;
 			i.start = inode->start_block;
+			i.block_start = start;
+			i.block_offset = offset;
 			i.sparse = 0;
-			i.block_ptr = block_ptr + sizeof(*inode);
 			break;
 		}	
 		case SQUASHFS_SYMLINK_TYPE: {
@@ -253,19 +295,25 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 
 			if(swap) {
 				squashfs_symlink_inode_header_2 sinodep;
-				memcpy(&sinodep, block_ptr, sizeof(sinodep));
-				SQUASHFS_SWAP_SYMLINK_INODE_HEADER_2(inodep,
-					&sinodep);
+				res = read_inode_data(&sinodep, &start, &offset, sizeof(sinodep));
+				if(res)
+					SQUASHFS_SWAP_SYMLINK_INODE_HEADER_2(inodep,
+						&sinodep);
 			} else
-				memcpy(inodep, block_ptr, sizeof(*inodep));
+				res = read_inode_data(inodep, &start, &offset, sizeof(*inodep));
+
+			if(res == FALSE)
+				EXIT_UNSQUASH("read_inode: failed to read "
+					"inode %lld:%d\n", start, offset);
 
 			i.symlink = malloc(inodep->symlink_size + 1);
 			if(i.symlink == NULL)
-				EXIT_UNSQUASH("read_inode: failed to malloc "
-					"symlink data\n");
-			strncpy(i.symlink, block_ptr +
-				sizeof(squashfs_symlink_inode_header_2),
-				inodep->symlink_size);
+				MEM_ERROR();
+
+			res = read_inode_data(i.symlink, &start, &offset, inodep->symlink_size);
+			if(res == FALSE)
+				EXIT_UNSQUASH("read_inode: failed to read "
+					"inode symbolic link %lld:%d\n", start, offset);
 			i.symlink[inodep->symlink_size] = '\0';
 			i.data = inodep->symlink_size;
 			break;
@@ -276,11 +324,16 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 
 			if(swap) {
 				squashfs_dev_inode_header_2 sinodep;
-				memcpy(&sinodep, block_ptr, sizeof(sinodep));
-				SQUASHFS_SWAP_DEV_INODE_HEADER_2(inodep,
-					&sinodep);
+				res = read_inode_data(&sinodep, &start, &offset, sizeof(sinodep));
+				if(res)
+					SQUASHFS_SWAP_DEV_INODE_HEADER_2(inodep,
+						&sinodep);
 			} else
-				memcpy(inodep, block_ptr, sizeof(*inodep));
+				res = read_inode_data(inodep, &start, &offset, sizeof(*inodep));
+
+			if(res == FALSE)
+				EXIT_UNSQUASH("read_inode: failed to read "
+					"inode %lld:%d\n", start, offset);
 
 			i.data = inodep->rdev;
 			break;
@@ -306,9 +359,9 @@ static struct dir *squashfs_opendir(unsigned int block_start, unsigned int offse
 		__attribute__((aligned));
 	squashfs_dir_entry_2 *dire = (squashfs_dir_entry_2 *) buffer;
 	long long start;
-	int bytes;
-	int dir_count, size;
-	struct dir_ent *new_dir;
+	int bytes = 0;
+	int dir_count, size, res;
+	struct dir_ent *ent, *cur_ent = NULL;
 	struct dir *dir;
 
 	TRACE("squashfs_opendir: inode start block %d, offset %d\n",
@@ -318,10 +371,10 @@ static struct dir *squashfs_opendir(unsigned int block_start, unsigned int offse
 
 	dir = malloc(sizeof(struct dir));
 	if(dir == NULL)
-		EXIT_UNSQUASH("squashfs_opendir: malloc failed!\n");
+		MEM_ERROR();
 
 	dir->dir_count = 0;
-	dir->cur_entry = 0;
+	dir->cur_entry = NULL;
 	dir->mode = (*i)->mode;
 	dir->uid = (*i)->uid;
 	dir->guid = (*i)->gid;
@@ -339,21 +392,20 @@ static struct dir *squashfs_opendir(unsigned int block_start, unsigned int offse
 		return dir;
 
 	start = sBlk.s.directory_table_start + (*i)->start;
-	bytes = lookup_entry(directory_table_hash, start);
-	if(bytes == -1)
-		EXIT_UNSQUASH("squashfs_opendir: directory block %d not "
-			"found!\n", block_start);
-
-	bytes += (*i)->offset;
+	offset = (*i)->offset;
 	size = (*i)->data + bytes;
 
 	while(bytes < size) {
 		if(swap) {
 			squashfs_dir_header_2 sdirh;
-			memcpy(&sdirh, directory_table + bytes, sizeof(sdirh));
-			SQUASHFS_SWAP_DIR_HEADER_2(&dirh, &sdirh);
+			res = read_directory_data(&sdirh, &start, &offset, sizeof(sdirh));
+			if(res)
+				SQUASHFS_SWAP_DIR_HEADER_2(&dirh, &sdirh);
 		} else
-			memcpy(&dirh, directory_table + bytes, sizeof(dirh));
+			res = read_directory_data(&dirh, &start, &offset, sizeof(dirh));
+
+		if(res == FALSE)
+			goto corrupted;
 
 		dir_count = dirh.count + 1;
 		TRACE("squashfs_opendir: Read directory header @ byte position "
@@ -369,12 +421,17 @@ static struct dir *squashfs_opendir(unsigned int block_start, unsigned int offse
 		while(dir_count--) {
 			if(swap) {
 				squashfs_dir_entry_2 sdire;
-				memcpy(&sdire, directory_table + bytes,
-					sizeof(sdire));
-				SQUASHFS_SWAP_DIR_ENTRY_2(dire, &sdire);
+				res = read_directory_data(&sdire, &start,
+					&offset, sizeof(sdire));
+				if(res)
+					SQUASHFS_SWAP_DIR_ENTRY_2(dire, &sdire);
 			} else
-				memcpy(dire, directory_table + bytes,
-					sizeof(*dire));
+				res = read_directory_data(dire, &start,
+					&offset, sizeof(*dire));
+
+			if(res == FALSE)
+				goto corrupted;
+
 			bytes += sizeof(*dire);
 
 			/* size should never be SQUASHFS_NAME_LEN or larger */
@@ -383,40 +440,63 @@ static struct dir *squashfs_opendir(unsigned int block_start, unsigned int offse
 				goto corrupted;
 			}
 
-			memcpy(dire->name, directory_table + bytes,
-				dire->size + 1);
+			res = read_directory_data(dire->name, &start, &offset,
+								dire->size + 1);
+
+			if(res == FALSE)
+				goto corrupted;
+
 			dire->name[dire->size + 1] = '\0';
+
+			/* check name for invalid characters (i.e /, ., ..) */
+			if(check_name(dire->name, dire->size + 1) == FALSE) {
+				ERROR("File system corrupted: invalid characters in name\n");
+				goto corrupted;
+			}
+
 			TRACE("squashfs_opendir: directory entry %s, inode "
 				"%d:%d, type %d\n", dire->name,
 				dirh.start_block, dire->offset, dire->type);
-			if((dir->dir_count % DIR_ENT_SIZE) == 0) {
-				new_dir = realloc(dir->dirs, (dir->dir_count +
-					DIR_ENT_SIZE) * sizeof(struct dir_ent));
-				if(new_dir == NULL)
-					EXIT_UNSQUASH("squashfs_opendir: "
-						"realloc failed!\n");
-				dir->dirs = new_dir;
-			}
-			strcpy(dir->dirs[dir->dir_count].name, dire->name);
-			dir->dirs[dir->dir_count].start_block =
-				dirh.start_block;
-			dir->dirs[dir->dir_count].offset = dire->offset;
-			dir->dirs[dir->dir_count].type = dire->type;
+
+			ent = malloc(sizeof(struct dir_ent));
+			if(ent == NULL)
+				MEM_ERROR();
+
+			ent->name = strdup(dire->name);
+			ent->start_block = dirh.start_block;
+			ent->offset = dire->offset;
+			ent->type = dire->type;
+			ent->next = NULL;
+			if(cur_ent == NULL)
+				dir->dirs = ent;
+			else
+				cur_ent->next = ent;
+			cur_ent = ent;
 			dir->dir_count ++;
 			bytes += dire->size + 1;
 		}
 	}
 
+	if(needs_sorting)
+		sort_directory(dir);
+
+	/* check directory for duplicate names and sorting */
+	if(check_directory(dir) == FALSE) {
+		if(needs_sorting)
+			ERROR("File system corrupted: directory has duplicate names\n");
+		else
+			ERROR("File system corrupted: directory has duplicate names or is unsorted\n");
+		goto corrupted;
+	}
 	return dir;
 
 corrupted:
-	free(dir->dirs);
-	free(dir);
+	squashfs_closedir(dir);
 	return NULL;
 }
 
 
-squashfs_operations *read_filesystem_tables_2()
+static int read_filesystem_tables()
 {
 	long long table_start;
 
@@ -487,37 +567,121 @@ squashfs_operations *read_filesystem_tables_2()
 		}
 	}
 
-	/* Read directory table */
-
-	/* Sanity check super block contents */
+	/* Sanity check super block directory table values */
 	if(sBlk.s.directory_table_start > table_start) {
 		ERROR("read_filesystem_tables: directory table start too large in super block\n");
 		goto corrupted;
 	}
 
-	directory_table = read_directory_table(sBlk.s.directory_table_start,
-				table_start);
-	if(directory_table == NULL)
-		goto corrupted;
-
-	/* Read inode table */
-
-	/* Sanity check super block contents */
+	/* Sanity check super block inode table values */
 	if(sBlk.s.inode_table_start >= sBlk.s.directory_table_start) {
 		ERROR("read_filesystem_tables: inode table start too large in super block\n");
 		goto corrupted;
 	}
 
-	inode_table = read_inode_table(sBlk.s.inode_table_start,
-				sBlk.s.directory_table_start);
-	if(inode_table == NULL)
-		goto corrupted;
-
-	return &ops;
+	return TRUE;
 
 corrupted:
-	ERROR("File system corruption detected\n");
-	return NULL;
+	return FALSE;
+}
+
+
+int read_super_2(squashfs_operations **s_ops, void *s)
+{
+	 squashfs_super_block_3 *sBlk_3 = s;
+
+	if(sBlk_3->s_magic != SQUASHFS_MAGIC || sBlk_3->s_major != 2 ||
+							sBlk_3->s_minor > 1)
+		return -1;
+
+	sBlk.s.s_magic = sBlk_3->s_magic;
+	sBlk.s.inodes = sBlk_3->inodes;
+	sBlk.s.mkfs_time = sBlk_3->mkfs_time;
+	sBlk.s.block_size = sBlk_3->block_size;
+	sBlk.s.fragments = sBlk_3->fragments;
+	sBlk.s.block_log = sBlk_3->block_log;
+	sBlk.s.flags = sBlk_3->flags;
+	sBlk.s.s_major = sBlk_3->s_major;
+	sBlk.s.s_minor = sBlk_3->s_minor;
+	sBlk.s.root_inode = sBlk_3->root_inode;
+	sBlk.s.bytes_used = sBlk_3->bytes_used_2;
+	sBlk.s.inode_table_start = sBlk_3->inode_table_start;
+	sBlk.s.directory_table_start = sBlk_3->directory_table_start_2;
+	sBlk.s.fragment_table_start = sBlk_3->fragment_table_start_2;
+	sBlk.s.inode_table_start = sBlk_3->inode_table_start_2;
+	sBlk.no_uids = sBlk_3->no_uids;
+	sBlk.no_guids = sBlk_3->no_guids;
+	sBlk.uid_start = sBlk_3->uid_start_2;
+	sBlk.guid_start = sBlk_3->guid_start_2;
+	sBlk.s.xattr_id_table_start = SQUASHFS_INVALID_BLK;
+
+	*s_ops = &ops;
+
+	/*
+	 * 2.x filesystems use gzip compression.
+	 */
+	comp = lookup_compressor("gzip");
+
+	if(sBlk_3->s_minor == 0)
+		needs_sorting = TRUE;
+
+	return TRUE;
+}
+
+
+static void squashfs_stat(char *source)
+{
+	time_t mkfs_time = (time_t) sBlk.s.mkfs_time;
+	struct tm *t = use_localtime ? localtime(&mkfs_time) :
+					gmtime(&mkfs_time);
+	char *mkfs_str = asctime(t);
+
+#if __BYTE_ORDER == __BIG_ENDIAN
+	printf("Found a valid %sSQUASHFS %d:%d superblock on %s.\n",
+		swap ? "little endian " : "big endian ", sBlk.s.s_major,
+		sBlk.s.s_minor, source);
+#else
+	printf("Found a valid %sSQUASHFS %d:%d superblock on %s.\n",
+		swap ? "big endian " : "little endian ", sBlk.s.s_major,
+		sBlk.s.s_minor, source);
+#endif
+
+	printf("Creation or last append time %s", mkfs_str ? mkfs_str :
+		"failed to get time\n");
+	printf("Filesystem size %llu bytes (%.2f Kbytes / %.2f Mbytes)\n",
+		sBlk.s.bytes_used, sBlk.s.bytes_used / 1024.0,
+		sBlk.s.bytes_used / (1024.0 * 1024.0));
+
+	printf("Block size %d\n", sBlk.s.block_size);
+	printf("Filesystem is %sexportable via NFS\n",
+		SQUASHFS_EXPORTABLE(sBlk.s.flags) ? "" : "not ");
+	printf("Inodes are %scompressed\n",
+		SQUASHFS_UNCOMPRESSED_INODES(sBlk.s.flags) ? "un" : "");
+	printf("Data is %scompressed\n",
+		SQUASHFS_UNCOMPRESSED_DATA(sBlk.s.flags) ? "un" : "");
+
+	if(SQUASHFS_NO_FRAGMENTS(sBlk.s.flags))
+		printf("Fragments are not stored\n");
+	else {
+		printf("Fragments are %scompressed\n",
+			SQUASHFS_UNCOMPRESSED_FRAGMENTS(sBlk.s.flags) ?  "un" : "");
+		printf("Always-use-fragments option is %sspecified\n",
+				SQUASHFS_ALWAYS_FRAGMENTS(sBlk.s.flags) ? "" : "not ");
+	}
+
+	printf("Check data is %spresent in the filesystem\n",
+		SQUASHFS_CHECK_DATA(sBlk.s.flags) ? "" : "not ");
+	printf("Duplicates are %sremoved\n", SQUASHFS_DUPLICATES(sBlk.s.flags) ? "" : "not ");
+	printf("Number of fragments %d\n", sBlk.s.fragments);
+	printf("Number of inodes %d\n", sBlk.s.inodes);
+	printf("Number of uids %d\n", sBlk.no_uids);
+	printf("Number of gids %d\n", sBlk.no_guids);
+
+	TRACE("sBlk.s.inode_table_start 0x%llx\n", sBlk.s.inode_table_start);
+	TRACE("sBlk.s.directory_table_start 0x%llx\n", sBlk.s.directory_table_start);
+	TRACE("sBlk.s.fragment_table_start 0x%llx\n\n", sBlk.s.fragment_table_start);
+	TRACE("sBlk.uid_start 0x%llx\n", sBlk.uid_start);
+	TRACE("sBlk.guid_start 0x%llx\n", sBlk.guid_start);
 }
 
 
@@ -525,5 +689,7 @@ static squashfs_operations ops = {
 	.opendir = squashfs_opendir,
 	.read_fragment = read_fragment,
 	.read_block_list = read_block_list,
-	.read_inode = read_inode
+	.read_inode = read_inode,
+	.read_filesystem_tables = read_filesystem_tables,
+	.stat = squashfs_stat
 };
