@@ -4,7 +4,7 @@
  * Create a squashfs filesystem.  This is a highly compressed read only
  * filesystem.
  *
- * Copyright (c) 2010, 2012, 2013, 2014, 2019, 2021
+ * Copyright (c) 2010, 2012, 2013, 2014, 2019, 2021, 2022
  * Phillip Lougher <phillip@squashfs.org.uk>
  *
  * This program is free software; you can redistribute it and/or
@@ -73,15 +73,17 @@ extern void save_xattrs();
 extern void restore_xattrs();
 extern unsigned int xattr_bytes, total_xattr_bytes;
 extern int write_xattr(char *, unsigned int);
-extern int read_xattrs_from_disk(int, struct squashfs_super_block *, int, long long *);
+extern unsigned int read_xattrs_from_disk(int, struct squashfs_super_block *, int, long long *);
 extern struct xattr_list *get_xattr(int, unsigned int *, int *);
 extern void free_xattr(struct xattr_list *, int);
 #else
+#include "squashfs_swap.h"
+
 static inline int get_xattrs(int fd, struct squashfs_super_block *sBlk)
 {
 	if(sBlk->xattr_id_table_start != SQUASHFS_INVALID_BLK) {
 		fprintf(stderr, "Xattrs in filesystem! These are not "
-			"supported on this version of Squashfs\n");
+			"supported on this build of Mksquashfs\n");
 		return 0;
 	} else
 		return SQUASHFS_INVALID_BLK;
@@ -116,14 +118,31 @@ static inline int write_xattr(char *pathname, unsigned int xattr)
 }
 
 
-static inline int read_xattrs_from_disk(int fd, struct squashfs_super_block *sBlk, int flag, long long *table_start)
+static inline unsigned int read_xattrs_from_disk(int fd, struct squashfs_super_block *sBlk, int sanity_only, long long *table_start)
 {
-	if(sBlk->xattr_id_table_start != SQUASHFS_INVALID_BLK) {
-		fprintf(stderr, "Xattrs in filesystem! These are not "
-			"supported on this version of Squashfs\n");
+	int res;
+	struct squashfs_xattr_table id_table;
+
+	/*
+	 * Read sufficient xattr metadata to obtain the start of the xattr
+	 * metadata on disk (table_start).  This value is needed to do
+	 * sanity checking of the filesystem.
+	 */
+	res = read_fs_bytes(fd, sBlk->xattr_id_table_start, sizeof(id_table), &id_table);
+	if(res == 0)
 		return 0;
-	} else
-		return SQUASHFS_INVALID_BLK;
+
+	SQUASHFS_INSWAP_XATTR_TABLE(&id_table);
+
+	/*
+	 * id_table.xattr_table_start stores the start of the compressed xattr
+	 * metadata blocks.  This by definition is also the end of the previous
+	 * filesystem table - the id lookup table.
+	 */
+	if(table_start != NULL)
+		*table_start = id_table.xattr_table_start;
+
+	return id_table.xattr_ids;
 }
 
 
